#version 420

#ifndef HDR_GLSL
#define HDR_GLSL

const float MAX_BRIGHTNESS = 30.0;
const float BRIGHTNESS_CURVE = 0.1;

vec4 encodeHDRColor(vec3 color) {
    color /= MAX_BRIGHTNESS;
    color = clamp(color, 0.0, 1.0);
    color = pow(color, vec3(BRIGHTNESS_CURVE));
    color *= vec3(2047.0, 2047.0, 1023.0);
    uvec3 rawOutput = uvec3(round(color)) << uvec3(0, 11, 22);
    uint result = rawOutput.x | rawOutput.y | rawOutput.z;
    return vec4(
        result & 255u,
        (result >> 8u) & 255u,
        (result >> 16u) & 255u,
        result >> 24u
    ) / 255.0;
}

vec3 decodeHDRColor(vec4 raw) {
    uvec4 scaled = uvec4(raw * 255.0) << uvec4(0, 8, 16, 24);
    uint data = scaled.x | scaled.y | scaled.z | scaled.w;
    vec3 color = vec3(
        float(data & 2047u),
        float((data >> 11u) & 2047u),
        float(data >> 22u)
    );
    color /= vec3(2047.0, 2047.0, 1023.0);
    color = pow(color, vec3(1.0 / BRIGHTNESS_CURVE));
    return color * MAX_BRIGHTNESS;
}

float luminance(vec3 color) {
    return dot(color, vec3(0.2125, 0.7154, 0.0721));
}

#endif // HDR_GLSL
#line 1 1010

struct MaterialMaskData {
    bool metal;
    float emission;
    vec3 normal;
};

const float MAX_VALUE = 16777215.0;

vec3 getMaskNormal(int data) {
    vec3 normal = vec3(
        (data >> 9) & 15, 
        (data >> 13) & 15, 
        0.0
    ) / 15.0 * 2.0 - 1.0;

    normal.z = sqrt(1.0 - min(1.0, dot(normal.xy, normal.xy)));
    normal.z *= float((data >> 17) & 1) * 2.0 - 1.0;    
    return normalize(normal);
}

vec3 getMaskNormal(float raw) {
    int data = int(round(raw * MAX_VALUE));
    return getMaskNormal(data);
}

MaterialMaskData getMaterialMask(float raw) {
    int data = int(round(raw * MAX_VALUE));

    return MaterialMaskData(
        (data & 1) == 1,
        ((data >> 1) & 255) / 255.0,
        getMaskNormal(data)
    );
}

float storeMaterialMask(MaterialMaskData data) {
    vec3 unitNormal = data.normal * 0.5 + 0.5;
    return (
        int(data.metal) | 
        (int(data.emission * 255.0) << 1) |
        (int(round(unitNormal.x * 15.0)) << 9) |
        (int(round(unitNormal.y * 15.0)) << 13) |
        (data.normal.z > 0.0 ? 1 : 0) << 17
    ) / MAX_VALUE;
}
#line 5 1020

in vec2 texCoord;

uniform sampler2D DiffuseSampler;
uniform sampler2D DepthSampler;
uniform sampler2D GISampler;
uniform sampler2D MaterialMask;

out vec4 fragColor;

// TONEMAPPING

vec3 ACESFilm(vec3 x)
{
    float a = 2.51;
    float b = 0.03;
    float c = 2.43;
    float d = 0.59;
    float e = 0.14;
    return clamp((x*(a*x+b))/(x*(c*x+d)+e), 0.0, 1.0);
}

void main() {
    vec3 gi = decodeHDRColor(texture(GISampler, texCoord));
    float depth = texture(DepthSampler, texCoord).r;

    vec3 color = gi;
    if (depth < 1.0) {
        MaterialMaskData maskData = getMaterialMask(texture(MaterialMask, texCoord).r);
        color = pow(texture(DiffuseSampler, texCoord).rgb, vec3(2.2));
        color = vec3(color) * (gi + maskData.emission * 10.0);
        color = ACESFilm(color);
    }
    
    color = pow(color, vec3(1.0 / 2.2));
    fragColor = vec4(color, 1.0);
}