#version 420

#ifndef HDR_GLSL
#define HDR_GLSL

const float MAX_BRIGHTNESS = 30.0;
const float BRIGHTNESS_CURVE = 0.1;

vec4 encodeHDRColor(vec3 color) {
    color /= MAX_BRIGHTNESS;
    color = clamp(color, 0.0, 1.0);
    color = pow(color, vec3(BRIGHTNESS_CURVE));
    color *= vec3(2047.0, 2047.0, 1023.0);
    uvec3 rawOutput = uvec3(round(color)) << uvec3(0, 11, 22);
    uint result = rawOutput.x | rawOutput.y | rawOutput.z;
    return vec4(
        result & 255u,
        (result >> 8u) & 255u,
        (result >> 16u) & 255u,
        result >> 24u
    ) / 255.0;
}

vec3 decodeHDRColor(vec4 raw) {
    uvec4 scaled = uvec4(raw * 255.0) << uvec4(0, 8, 16, 24);
    uint data = scaled.x | scaled.y | scaled.z | scaled.w;
    vec3 color = vec3(
        float(data & 2047u),
        float((data >> 11u) & 2047u),
        float(data >> 22u)
    );
    color /= vec3(2047.0, 2047.0, 1023.0);
    color = pow(color, vec3(1.0 / BRIGHTNESS_CURVE));
    return color * MAX_BRIGHTNESS;
}

float luminance(vec3 color) {
    return dot(color, vec3(0.2125, 0.7154, 0.0721));
}

#endif // HDR_GLSL
#ifndef CONSTANTS_GLSL
#define CONSTANTS_GLSL

// Application specific
const ivec2 GRID_SIZE = ivec2(1024, 705);
const int AREA_SIDE_LENGTH = int(pow(float(GRID_SIZE.x * GRID_SIZE.y / 2), 1.0 / 3.0));
const int STORAGE_SIZE = 839;
const uint MAX_COUNTER = 255u;

// Universal
const float EPSILON = 0.001;
const float PI = 3.141592654;

// Temporal

const float MAX_TEMPORAL_BLENDING = 0.05;
const uint SDF_GENERATION_FRAMES = 6u;
const float MAX_SDF_DIST = 255.0;

#endif // CONSTANTS_GLSL
#ifndef RAND_GLSL
#define RAND_GLSL

uint state;

uint rand() {
	uint newState = state * uint(747796405) + uint(2891336453);
	uint word = ((newState >> ((newState >> uint(28)) + uint(4))) ^ newState) * uint(277803737);
    state = (word >> uint(22)) ^ word;
    return state;
}

float randFloat() {
    return float(rand() & uvec3(0x7fffffffU)) / float(0x7fffffff);
}

vec2 randVec2() {
    return vec2(randFloat(), randFloat());
}

vec3 randVec3() {
    return vec3(randFloat(), randFloat(), randFloat());
}

vec4 randVec4() {
    return vec4(randFloat(), randFloat(), randFloat(), randFloat());
}

void initRNG(uvec2 pixel, uvec2 resolution, uint frame) {
    state = frame;
    state = (pixel.x + pixel.y * resolution.x) ^ rand();
    rand();
}

#endif // RAND_GLSL
#ifndef VOXELIZATION
#define VOXELIZATION

#ifndef CONSTANTS_GLSL
#define CONSTANTS_GLSL

// Application specific
const ivec2 GRID_SIZE = ivec2(1024, 705);
const int AREA_SIDE_LENGTH = int(pow(float(GRID_SIZE.x * GRID_SIZE.y / 2), 1.0 / 3.0));
const int STORAGE_SIZE = 839;
const uint MAX_COUNTER = 255u;

// Universal
const float EPSILON = 0.001;
const float PI = 3.141592654;

// Temporal

const float MAX_TEMPORAL_BLENDING = 0.05;
const uint SDF_GENERATION_FRAMES = 6u;
const float MAX_SDF_DIST = 255.0;

#endif // CONSTANTS_GLSL

ivec2 positionToCellScreen(vec3 position, out bool inside) {
    ivec3 sides = ivec3(AREA_SIDE_LENGTH);

    ivec3 iPosition = ivec3(floor(position));
    iPosition += sides / 2;

    inside = true;
    if (clamp(iPosition, ivec3(0), sides - 1) != iPosition) {
        inside = false;
        return ivec2(-1);
    }

    int index = (iPosition.y * sides.z + iPosition.z) * sides.x + iPosition.x;
    
    int halfWidth = GRID_SIZE.x / 2;
    ivec2 result = ivec2(
        (index % halfWidth) * 2,
        index / halfWidth + 1
    );
    result.x += result.y % 2;

    return result;
}

ivec2 cellToPixelScreen(ivec2 cell, ivec2 screenSize) {
    return ivec2(round(vec2(cell) / GRID_SIZE * screenSize));
}

ivec2 positionToPixel(vec3 position, out bool inside) {
    ivec3 sides = ivec3(AREA_SIDE_LENGTH);

    ivec3 iPosition = ivec3(floor(position));
    iPosition += sides / 2;

    inside = true;
    if (clamp(iPosition, ivec3(0), sides - 1) != iPosition) {
        inside = false;
        return ivec2(-1);
    }

    int index = (iPosition.y * sides.z + iPosition.z) * sides.x + iPosition.x;
    
    return ivec2(
        index % STORAGE_SIZE,
        index / STORAGE_SIZE
    );
}

ivec3 pixelToPosition(ivec2 pixel) {
    ivec3 sides = ivec3(AREA_SIDE_LENGTH);

    int index = pixel.x + pixel.y * STORAGE_SIZE;
    return ivec3(
        index % sides.x,
        index / sides.x / sides.z,
        (index / sides.x) % sides.z
    ) - sides / 2;
}

#endif // VOXELIZATION
#ifndef CONVERSIONS
#define CONVERSIONS

vec3 screenToView(vec3 screenPos, mat4 projInv) {
    vec4 ndc = vec4(screenPos * 2.0 - 1.0, 1.0);
    vec4 viewPos = projInv * ndc;
    return viewPos.xyz / viewPos.w;
}

vec3 viewToScreen(vec3 viewPos, mat4 projMat) {
    vec4 clipEnd = projMat * vec4(viewPos, 1);
    return clipEnd.xyz / clipEnd.w * 0.5 + 0.5;
}

float linearizeDepth(float depth, float near, float far) {
    return (near * far) / (depth * (near - far) + far);
}

#endif // CONVERSIONS
#line 1 1010

struct MaterialMaskData {
    bool metal;
    float emission;
    vec3 normal;
};

const float MAX_VALUE = 16777215.0;

vec3 getMaskNormal(int data) {
    vec3 normal = vec3(
        (data >> 9) & 15, 
        (data >> 13) & 15, 
        0.0
    ) / 15.0 * 2.0 - 1.0;

    normal.z = sqrt(1.0 - min(1.0, dot(normal.xy, normal.xy)));
    normal.z *= float((data >> 17) & 1) * 2.0 - 1.0;    
    return normalize(normal);
}

vec3 getMaskNormal(float raw) {
    int data = int(round(raw * MAX_VALUE));
    return getMaskNormal(data);
}

MaterialMaskData getMaterialMask(float raw) {
    int data = int(round(raw * MAX_VALUE));

    return MaterialMaskData(
        (data & 1) == 1,
        ((data >> 1) & 255) / 255.0,
        getMaskNormal(data)
    );
}

float storeMaterialMask(MaterialMaskData data) {
    vec3 unitNormal = data.normal * 0.5 + 0.5;
    return (
        int(data.metal) | 
        (int(data.emission * 255.0) << 1) |
        (int(round(unitNormal.x * 15.0)) << 9) |
        (int(round(unitNormal.y * 15.0)) << 13) |
        (data.normal.z > 0.0 ? 1 : 0) << 17
    ) / MAX_VALUE;
}
#ifndef QUATERNIONS_GLSL
#define QUATERNIONS_GLSL

#define quat vec4

quat quaternionMultiply(quat a, quat b) {
    return vec4(a.w * b.xyz + b.w * a.xyz + cross(a.xyz, b.xyz), a.w * b.w - dot(a.xyz, b.xyz));
}

vec3 quaternionRotate(vec3 pos, vec3 axis, float angle) {
    vec4 q = vec4(sin(angle / 2.0) * axis, cos(angle / 2.0));
    vec4 qInv = vec4(-q.xyz, q.w);
    return quaternionMultiply(quaternionMultiply(q, vec4(pos, 0)), qInv).xyz;
}

vec3 quaternionRotate(vec3 pos, quat q) {
    vec4 qInv = vec4(-q.xyz, q.w);
    vec4 q1 = quaternionMultiply(q, vec4(pos, 0));
    return q1.w * qInv.xyz + qInv.w * q1.xyz + cross(q1.xyz, qInv.xyz);
}

quat quaternionInverse(quat q) {
    return quat(-q.xyz, q.w);
}

quat getRotationToZAxis(vec3 vec) {
	if (vec.z < -0.99999) return quat(1.0, 0.0, 0.0, 0.0);

	return normalize(quat(vec.y, -vec.x, 0.0, 1.0 + vec.z));
}


#endif // QUATERNIONS_GLSL
#ifndef BRDF_GLSL
#define BRDF_GLSL

#ifndef CONSTANTS_GLSL
#define CONSTANTS_GLSL

// Application specific
const ivec2 GRID_SIZE = ivec2(1024, 705);
const int AREA_SIDE_LENGTH = int(pow(float(GRID_SIZE.x * GRID_SIZE.y / 2), 1.0 / 3.0));
const int STORAGE_SIZE = 839;
const uint MAX_COUNTER = 255u;

// Universal
const float EPSILON = 0.001;
const float PI = 3.141592654;

// Temporal

const float MAX_TEMPORAL_BLENDING = 0.05;
const uint SDF_GENERATION_FRAMES = 6u;
const float MAX_SDF_DIST = 255.0;

#endif // CONSTANTS_GLSL
#ifndef QUATERNIONS_GLSL
#define QUATERNIONS_GLSL

#define quat vec4

quat quaternionMultiply(quat a, quat b) {
    return vec4(a.w * b.xyz + b.w * a.xyz + cross(a.xyz, b.xyz), a.w * b.w - dot(a.xyz, b.xyz));
}

vec3 quaternionRotate(vec3 pos, vec3 axis, float angle) {
    vec4 q = vec4(sin(angle / 2.0) * axis, cos(angle / 2.0));
    vec4 qInv = vec4(-q.xyz, q.w);
    return quaternionMultiply(quaternionMultiply(q, vec4(pos, 0)), qInv).xyz;
}

vec3 quaternionRotate(vec3 pos, quat q) {
    vec4 qInv = vec4(-q.xyz, q.w);
    vec4 q1 = quaternionMultiply(q, vec4(pos, 0));
    return q1.w * qInv.xyz + qInv.w * q1.xyz + cross(q1.xyz, qInv.xyz);
}

quat quaternionInverse(quat q) {
    return quat(-q.xyz, q.w);
}

quat getRotationToZAxis(vec3 vec) {
	if (vec.z < -0.99999) return quat(1.0, 0.0, 0.0, 0.0);

	return normalize(quat(vec.y, -vec.x, 0.0, 1.0 + vec.z));
}


#endif // QUATERNIONS_GLSL
#ifndef RAND_GLSL
#define RAND_GLSL

uint state;

uint rand() {
	uint newState = state * uint(747796405) + uint(2891336453);
	uint word = ((newState >> ((newState >> uint(28)) + uint(4))) ^ newState) * uint(277803737);
    state = (word >> uint(22)) ^ word;
    return state;
}

float randFloat() {
    return float(rand() & uvec3(0x7fffffffU)) / float(0x7fffffff);
}

vec2 randVec2() {
    return vec2(randFloat(), randFloat());
}

vec3 randVec3() {
    return vec3(randFloat(), randFloat(), randFloat());
}

vec4 randVec4() {
    return vec4(randFloat(), randFloat(), randFloat(), randFloat());
}

void initRNG(uvec2 pixel, uvec2 resolution, uint frame) {
    state = frame;
    state = (pixel.x + pixel.y * resolution.x) ^ rand();
    rand();
}

#endif // RAND_GLSL

const vec3 PREDETERMINED_F0[] = vec3[](
    vec3(0.53123, 0.51236, 0.49583), // Iron
    vec3(0.94423, 0.77610, 0.37340), // Gold
    vec3(0.91230, 0.91385, 0.91968), // Aluminium
    vec3(0.55560, 0.55454, 0.55478), // Chrome
    vec3(0.92595, 0.72090, 0.50415), // Copper
    vec3(0.63248, 0.62594, 0.64148), // Lead
    vec3(0.67885, 0.64240, 0.58841), // Platinum
    vec3(0.96200, 0.94947, 0.92212)  // Silver
);

struct Material {
    vec3 albedo;
    vec3 normal;
    float ambientOcclusion;
    float roughness;
    vec3 F0;
    float emission;
    bool metal;
};

Material defaultMaterial(vec3 normal) {
    return Material(
        vec3(1),
        normal,
        1.0,
        0.0,
        vec3(0.04),
        0.0,
        false
    );
}

vec3 cosineWeighted(vec3 normal) {
    vec2 v = randVec2();
    float angle = 2.0 * PI * v.x;
    float u = 2.0 * v.y - 1.0;

    vec3 directionOffset = vec3(sqrt(1.0 - u * u) * vec2(cos(angle), sin(angle)), u);
    return normalize(normal + directionOffset);
}

vec3 fresnelSchlick(vec3 F0, float cosTheta) {
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}

vec3 diffuseHammon(float LdotV, float NdotL, float NdotV, float NdotH, Material material) {
    float facing = 0.5 + 0.5 * LdotV;
    float rough = facing * (0.9 - 0.4 * facing) * (0.5 + NdotH) / NdotH;
    float smoot = 1.05 * (1.0 - pow(1.0 - NdotL, 5.0)) * (1.0 - pow(1.0 - NdotV, 5.0));
    float single = 1.0 / PI * mix(smoot, rough, material.roughness);
    float multi = 0.1159 * material.roughness;
    return material.albedo * (single + material.albedo * multi);
}

float ggxTrowbridgeReitz(float NdotH, Material material) {
    float denom = NdotH * NdotH * (material.roughness * material.roughness - 1.0) + 1.0;
    return material.roughness * material.roughness / (PI * denom * denom);
}

float ggxG1Schlick(float cosTheta, float k) {
    float denom = cosTheta * (1.0 - k) + k;
    return cosTheta / denom;
}

float geometrySmith(float NdotV, float NdotL, Material material) {
    float k = (material.roughness + 1.0) * (material.roughness + 1.0) / 8.0;
    return ggxG1Schlick(NdotV, k) * ggxG1Schlick(NdotL, k);
}

vec3 cookTorrance(float NdotH, float NdotV, float NdotL, Material material) {
    return ggxTrowbridgeReitz(NdotH, material) * 
            geometrySmith(NdotV, NdotL, material) * 
            fresnelSchlick(material.F0, NdotV) / (4.0 * NdotV * NdotL);
}

vec3 brdfDirect(vec3 view, vec3 light, Material material) {
    vec3 halfway = normalize(view + light);
    float NdotL = max(dot(material.normal, light), 0.0);
    float NdotV = max(dot(material.normal, view), 0.0);
    float NdotH = max(dot(material.normal, halfway), 0.0);
    float LdotV = dot(light, view);
    vec3 diffuse = diffuseHammon(LdotV, NdotL, NdotV, NdotH, material);
    vec3 specular = cookTorrance(NdotH, NdotV, NdotL, material);
    return (max(specular * mix(vec3(1.0), material.albedo, float(material.metal)), 0.0) + diffuse * (1.0 - float(material.metal))) * NdotL;
}

float evaluateSpecularProbability(vec3 view, Material material) {
    vec3 reflected = reflect(-view, material.normal);
    vec3 halfway = normalize(view + reflected);
    float fresnel = clamp(luminance(fresnelSchlick(material.F0, max(dot(view, halfway), 0.0))), 0.0, 1.0);
    
    float diffuse = luminance(material.albedo) * (1.0 - float(material.metal)) * (1.0 - fresnel);
    float specular = fresnel;
    
    return specular / (specular + diffuse);
}

vec3 sampleGGXVNDF(vec3 Ve, vec2 alpha2D) {
    vec2 u = randVec2();

	vec3 Vh = normalize(vec3(alpha2D.x * Ve.x, alpha2D.y * Ve.y, Ve.z));

	float lensq = Vh.x * Vh.x + Vh.y * Vh.y;
	vec3 T1 = lensq > 0.0f ? vec3(-Vh.y, Vh.x, 0.0f) * inversesqrt(lensq) : vec3(1.0f, 0.0f, 0.0f);
	vec3 T2 = cross(Vh, T1);

	float r = sqrt(u.x);
	float phi = 2.0 * PI * u.y;
	float t1 = r * cos(phi);
	float t2 = r * sin(phi);
	float s = 0.5f * (1.0f + Vh.z);
	t2 = mix(sqrt(1.0f - t1 * t1), t2, s);

	vec3 Nh = t1 * T1 + t2 * T2 + sqrt(max(0.0f, 1.0f - t1 * t1 - t2 * t2)) * Vh;

	return normalize(vec3(alpha2D.x * Nh.x, alpha2D.y * Nh.y, max(0.0f, Nh.z)));
}

vec3 sampleSpecular(vec3 view, Material material, out vec3 rayDirection) {
    float alpha = material.roughness * material.roughness;
    vec3 microfacetNormal;
    if (material.roughness > EPSILON) {
        quat qRotation = getRotationToZAxis(material.normal);
        microfacetNormal = normalize(sampleGGXVNDF(quaternionRotate(view, qRotation), vec2(alpha)));
        microfacetNormal = quaternionRotate(microfacetNormal, vec4(qRotation.xyz, -qRotation.w));
    } else {
        microfacetNormal = material.normal;
    }
    rayDirection = reflect(-view, microfacetNormal);
    vec3 halfway = normalize(view + rayDirection);
    
    float HdotL = max(dot(halfway, rayDirection), 0.0);
    float NdotL = max(dot(microfacetNormal, rayDirection), 0.0);
    float NdotV = max(dot(microfacetNormal, view), 0.0);
    float k = (alpha + 1.0);
    k *= k / 8.0;

    vec3 F = fresnelSchlick(material.F0, NdotV);
    float G2 = ggxG1Schlick(NdotL, k);
    
    return F * G2;
}

vec3 sampleDiffuse(vec3 view, Material material, out vec3 rayDirection) {
    rayDirection = cosineWeighted(material.normal);
    vec3 halfway = normalize(view + rayDirection);

    float NdotL = max(dot(material.normal, rayDirection), 0.0);
    float NdotV = max(dot(material.normal, view), 0.0);
    float NdotH = max(dot(material.normal, halfway), 0.0);
    float LdotV = dot(rayDirection, view);

    return diffuseHammon(LdotV, NdotL, NdotV, NdotH, material) / NdotL * PI;
}

#endif // BRDF_GLSL
#line 11 0

in vec2 texCoord;
in vec3 sunDir;
in mat4 projMat;
in mat3 modelViewMat;
in vec3 chunkOffset;
in vec3 rayDir;
in float near;
in float far;
in mat4 projInv;
flat in uint frame;
in float renderScale;
in vec3 steveDirection;
flat in ivec2 atlasSize;

uniform sampler2D DiffuseSampler;
uniform sampler2D DepthSampler;
uniform sampler2D NormalSampler;

uniform sampler2D VoxelSampler;
uniform sampler2D VoxelDepthSampler;

uniform sampler2D Atlas;
uniform sampler2D Models;
uniform sampler2D SteveSampler;

uniform vec2 InSize;

out vec4 fragColor;

const float SUN_SIZE_FACTOR = 0.9999;
const float SUN_INTENSITY = 5.0;
const float EMISSION_STRENGTH = 30.0;
const float WATER_F0 = (1.333 - 1.0) * (1.333 - 1.0) / (1.333 + 1.0) / (1.333 + 1.0);

const float GAMMA_CORRECTION = 2.2;

const float ANGLE_22_5 = 0.39269908169872414;

struct Ray {
    vec3 origin;
    ivec3 cell;
    vec3 direction;
};

// MAIN 

vec2 boxIntersection( in vec3 ro, in vec3 rd, vec3 boxSize, out vec3 outNormal ) 
{
    vec3 m = 1.0/rd; // can precompute if traversing a set of aligned boxes
    vec3 n = m*ro;   // can precompute if traversing a set of aligned boxes
    vec3 k = abs(m)*boxSize;
    vec3 t1 = -n - k;
    vec3 t2 = -n + k;
    float tN = max( max( t1.x, t1.y ), t1.z );
    float tF = min( min( t2.x, t2.y ), t2.z );
    if( tN>tF || tF<0.0) return vec2(-1.0); // no intersection
    outNormal = (tN>=0.0) ? step(vec3(tN),t1) : // ro ouside the box
                           step(t2,vec3(tF));  // ro inside the box
    outNormal *= -sign(rd);
    return vec2( tN, tF );
}

int getSDF(ivec2 pixel) {
    float depth = texelFetch(VoxelDepthSampler, pixel, 0).r;
    if (depth == 1.0)
        return 1;
    return int(depth * MAX_SDF_DIST);
}

struct Intersection {
    bool hit;
    int skip;
    mat3 tbn;
    vec2 uv;
    float dist;
    ivec2 pixel;
};

Intersection noIntersection(int skip) {
    Intersection res;
    res.hit = false;
    res.skip = skip;
    return res;
}

Intersection checkIntersection(ivec2 pixel, Ray ray) {
    ivec3 raw = ivec3(texelFetch(VoxelSampler, pixel, 0).rgb * 255.0);
    int index = raw.r | (raw.g << 8) | (raw.b << 16);
    int count = int(texelFetch(Models, ivec2(0, index), 0).r * 255.0);
    
    float minT = -1.0;
    vec2 uv;
    vec3 tangent;
    vec3 normal;
    for (int i = 0; i < count; i++) {
        const int bytesPerModel = 16;
        
        vec3 from = texelFetch(Models, ivec2(bytesPerModel * i + 1, index), 0).rgb * 3.0 - 1.0;
        vec3 to = texelFetch(Models, ivec2(bytesPerModel * i + 2, index), 0).rgb * 3.0 - 1.0;
        vec4 rotation = texelFetch(Models, ivec2(bytesPerModel * i + 3, index), 0);
        vec3 pivot = texelFetch(Models, ivec2(bytesPerModel * i + 4, index), 0).rgb;
        
        float angle = (rotation.w * 255.0 - 2.0) * ANGLE_22_5;
        vec4 quaternion = vec4(
            -rotation.xyz * sin(angle / 2.0),
            cos(angle / 2.0)
        );
        
        vec3 size = (to - from) / 2.0;
        vec3 center = (from + to) / 2.0;
        vec3 newNormal;
        
        vec3 ro = ray.origin - ray.cell - center - ray.direction * 0.2;
        vec3 rd = ray.direction;
        
        ro -= pivot - center;
        ro = quaternionRotate(ro, quaternion);
        rd = quaternionRotate(rd, quaternion);
        ro += pivot - center;

        vec2 t = boxIntersection(ro, rd, size, newNormal);    
        if (minT < 0.0 || (t.x >= 0.0 && t.x < minT)) {
            vec3 hitPos = ro + rd * t.x;
            if (clamp(hitPos, -0.501, 0.501) != hitPos) continue;
            vec3 relativePos = (hitPos + size) / size / 2.0;
            int faceID;
            vec2 newUV;
            vec3 newTangent;
            if (newNormal.x > 0.99) {
                faceID = 4;
                newUV = vec2(1.0 - relativePos.z, 1.0 - relativePos.y);
                newTangent = vec3(0.0, 0.0, -1.0); 
            } else if (newNormal.x < -0.99) {
                faceID = 5;
                newUV = vec2(relativePos.z, 1.0 - relativePos.y);
                newTangent = vec3(0.0, 0.0, 1.0);
            } else if (newNormal.y > 0.99) {
                faceID = 1;
                newUV = vec2(relativePos.x, relativePos.z);
                newTangent = vec3(1.0, 0.0, 0.0);
            } else if (newNormal.y < -0.99) {
                faceID = 0;
                newUV = vec2(1.0 - relativePos.x, relativePos.z);
                newTangent = vec3(-1.0, 0.0, 0.0);
            } else if (newNormal.z > 0.99) {
                faceID = 2;
                newUV = vec2(relativePos.x, 1.0 - relativePos.y);
                newTangent = vec3(1.0, 0.0, 0.0);
            } else if (newNormal.z < -0.99) {
                faceID = 3;
                newUV = vec2(1.0 - relativePos.x, 1.0 - relativePos.y);
                newTangent = vec3(-1.0, 0.0, 0.0);
            }
            ivec4 textureData = ivec4(texelFetch(Models, ivec2(bytesPerModel * i + 5 + faceID * 2, index), 0) * 255.0);            
            if (textureData.a > 200)
                continue;
                
            vec4 uvData = texelFetch(Models, ivec2(bytesPerModel * i + 5 + faceID * 2 + 1, index), 0);
            int textureIndex = textureData.r | (textureData.g << 8) | (textureData.b << 16);
            vec2 topLeft = vec2(textureIndex & 1023, (textureIndex >> 10) & 1023);
            float size = exp2(float(textureIndex >> 20));
            
            newUV = (mix(uvData.xy, uvData.zw, newUV) * size + topLeft) * 16.0;
            
            if (texelFetch(Atlas, ivec2(newUV), 0).a < 0.5)
                continue;

            minT = t.x;
            uv = newUV;

            vec4 invQuaternion = vec4(-quaternion.xyz, quaternion.w);
            newNormal = quaternionRotate(newNormal, invQuaternion);
            newTangent = quaternionRotate(newTangent, invQuaternion);
            normal = newNormal;
            tangent = newTangent;
        }
    }

    mat3 tbn = mat3(tangent, cross(normal, tangent), normal);
    return Intersection(minT > 0.0, 0, tbn, uv, minT - 0.2, pixel);
}

Intersection checkIntersection(Ray ray) {
    bool inside;
    ivec2 pixel = positionToPixel(vec3(ray.cell), inside);

    int skip = 0;
    if (!inside || (skip = getSDF(pixel)) != 0)
        return noIntersection(skip);

    return checkIntersection(pixel, ray);
}

mat3 getTBN(vec3 normal) {
    float xSign = abs(normal.x) < 0.1 ? 0.0 : sign(normal.x);
    float ySign = abs(normal.y) < 0.1 ? 0.0 : sign(normal.y);
    float zSign = abs(normal.z) < 0.1 ? 0.0 : sign(normal.z);
    vec3 tangent = vec3(zSign - ySign, 0, -xSign);
    vec3 bitangent = cross(tangent, normal);
    return mat3(tangent, bitangent, normal);
}

Material getMaterial(Intersection intersection) {
    uvec4 dataRaw = uvec4(round(texelFetch(VoxelSampler, intersection.pixel, 0) * 255.0)) << uvec4(0u, 8u, 16u, 0u);
    uint data = dataRaw.r | dataRaw.g | dataRaw.b;
    vec3 tintColor = vec3(
        float(dataRaw.a & 3u) / 3.0,
        float((dataRaw.a >> 2u) & 7u) / 7.0,
        float(dataRaw.a >> 5u) / 3.0
    );
    
    ivec2 atlasTexCoord = ivec2(intersection.uv);
    
    vec3 albedo = pow(texelFetch(Atlas, atlasTexCoord, 0).rgb * tintColor, vec3(GAMMA_CORRECTION));
    vec3 normalData = texelFetch(Atlas, atlasTexCoord + ivec2(atlasSize.x, 0), 0).rgb;
    vec3 normal;
    float ambientOcclusion;
    if (dot(normalData.xy, normalData.xy) < 0.01) {
        normal = intersection.tbn[2];
        ambientOcclusion = 1.0;
    } else {
        normalData.xy = normalData.xy * 2.0 - 1.0;
        normal = intersection.tbn * normalize(vec3(normalData.xy, sqrt(1.0 - dot(normalData.xy, normalData.xy)))),
        ambientOcclusion = normalData.b;
    }
    
    vec4 specularData = texelFetch(Atlas, atlasTexCoord + ivec2(0, atlasSize.y), 0);
    
    vec3 F0 = vec3(specularData.g);
    int index = int(round(specularData.g * 255.0));
    bool metal = index >= 230;
    if (index == 255) {
        F0 = albedo;
        albedo = vec3(1);
    } else if (metal) {
        F0 = PREDETERMINED_F0[index - 230];
    }
    
    return Material(
        albedo,
        normal,
        ambientOcclusion,
        (1.0 - specularData.r) * (1.0 - specularData.r),
        F0,
        specularData.a == 1.0 ? 0.0 : specularData.a,
        metal
    );
}

struct Hit {
    bool hit;
    float t;
    vec3 geometryNormal;
    Material material;
};

Hit noHit() {
    Hit hit;
    hit.hit = false;
    return hit;
}

struct PathTraceOutput {
    vec3 radiance;
    MaterialMaskData materialMask;
};

vec2 getTexCoord(vec3 pos, vec3 normal) {
    vec3 absNormal = abs(normal);
    float maxNormal = max(absNormal.x, max(absNormal.y, absNormal.z));
    vec3 flatNormal = vec3(greaterThanEqual(absNormal, vec3(maxNormal))) * sign(normal);
    vec3 fractPos = fract(pos);
    return max(flatNormal.x, 0.0) * vec2(1.0 - fractPos.z, 1.0 - fractPos.y) +
        max(-flatNormal.x, 0.0) * vec2(fractPos.z, 1.0 - fractPos.y) +
        max(flatNormal.y, 0.0) * vec2(fractPos.x, fractPos.z) +
        max(-flatNormal.y, 0.0) * vec2(fractPos.x, 1.0 - fractPos.z) +
        max(flatNormal.z, 0.0) * vec2(fractPos.x, 1.0 - fractPos.y) +
        max(-flatNormal.z, 0.0) * vec2(1.0 - fractPos.x, 1.0 - fractPos.y);
}

float traceSteve(Ray ray, vec3 point, vec3 normal, out vec2 uv) {
    float t = dot(point - ray.origin, normal) / dot(ray.direction, normal);
    vec3 hitPoint = ray.origin + ray.direction * t;
    vec3 bitangent = vec3(0, 1, 0);
    vec3 tangent = normalize(cross(bitangent, normal));
    uv = vec2(dot(tangent, hitPoint - point), dot(-bitangent, hitPoint - point));
    uv *= vec2(1.0, 22.0 / 39.0) * 1.875;
    uv = uv * 0.5 + 0.5 - vec2(0, 0.35);
    uv.x = 1.0 - uv.x;
    if (clamp(uv, vec2(0.0), vec2(1.0)) != uv)
        return -1.0;
    return t;
}

Hit raytrace(inout Ray ray, bool checkSteve, int maxSteps) {
    Intersection intersection = checkIntersection(ray);
    if (intersection.dist > 0.01) {
        vec3 normal = intersection.tbn[2];
        ray.origin += normal * 0.01;
        
        Hit hit = Hit(true, intersection.dist, normal, getMaterial(intersection));
        return hit;
    }

    // Steve
    vec2 steveUV;
    float steveDist = checkSteve ? traceSteve(ray, -fract(chunkOffset), steveDirection, steveUV) : -1.0;
    vec4 steveColor = checkSteve ? texture(SteveSampler, steveUV) : vec4(0.0);
    
    // Voxel tracing
    vec3 nextEdge = max(sign(ray.direction), 0.0);
    vec3 steps = (nextEdge - fract(ray.origin)) / ray.direction;
    vec3 originalStepSizes = abs(1.0 / ray.direction);
    vec3 rdSign = sign(ray.direction);
    
    float dist = 0.0;
    int toSkip = 0;
    for (int i = 0; i < maxSteps; i++) {
        float stepSize = min(steps.x, min(steps.y, steps.z));
        ray.origin += ray.direction * stepSize;
        vec3 stepAxis = vec3(lessThanEqual(steps, vec3(stepSize)));
        ivec3 nextCell = ray.cell + ivec3(stepAxis * rdSign);
        dist += stepSize;
        
        if (steveDist > 0.0 && dist > steveDist && steveColor.a > 0.0) {
            ray.origin += ray.direction * steveDist + steveDirection * 0.01;
            ray.cell = ivec3(floor(ray.origin));
            return Hit(
                true, 
                steveDist, 
                steveDirection, 
                Material(
                    steveColor.rgb,
                    steveDirection,
                    0.0,
                    0.5,
                    vec3(0.04),
                    0.0,
                    false
                )
            );
        }
        
        if (toSkip <= 0) {
            Intersection intersection = checkIntersection(Ray(ray.origin, nextCell, ray.direction));
            if (intersection.hit) {
                vec3 normal = intersection.tbn[2];
                ray.origin += normal * 0.001;
                
                Hit hit = Hit(true, dist + intersection.dist, normal, getMaterial(intersection));
                return hit;
            }
            toSkip = intersection.skip;
        }
        toSkip--;

        ray.cell = nextCell;
        steps += originalStepSizes * stepAxis - stepSize;
    }
    return noHit();
}

vec3 fresnel(vec3 F0, float cosTheta) {
    return F0 + (1.0 - F0) * pow(max(1.0 - cosTheta, 0.0), 5);
}

bool checkSun(vec3 hitPos, ivec3 hitCell, int maxSteps, vec3 normal, out vec3 dir) {
    vec3 lightDir = sunDir;
    if (lightDir.y < -0.2)
        lightDir *= -1.0;
    dir = mix(cosineWeighted(lightDir), lightDir, SUN_SIZE_FACTOR);
    if (dot(dir, normal) <= 0.0)
        return false;
    Ray sunRay = Ray(hitPos, hitCell, dir);
    return !raytrace(sunRay, true, maxSteps).hit;
}

vec3 getNextRayDirection(Material material, vec3 rayDir, inout vec3 throughput) {
    vec3 nextRayDir;
    float specularProbability = evaluateSpecularProbability(-rayDir, material);
    if (randFloat() < specularProbability) {
        throughput /= specularProbability;
        throughput *= sampleSpecular(-rayDir, material, nextRayDir) * mix(vec3(1.0), material.albedo, float(material.metal));
    } else {
        throughput /= 1.0 - specularProbability;
        throughput *= sampleDiffuse(-rayDir, material, nextRayDir);
    }
    return nextRayDir;
}

const vec3 TOP_COLOR = vec3(0.017, 0.011, 0.773);
const vec3 BOTTOM_COLOR = pow(vec3(0.5, 0.52, 0.87), vec3(2.2));
const vec3 FOG_COLOR = pow(vec3(0.69, 0.7, 0.86), vec3(2.2));
const vec3 GROUND_COLOR = pow(vec3(0.33, 0.31, 0.29), vec3(2.2));
const vec3 SUNSET_COLOR = pow(vec3(1.2, 0.18, 0), vec3(2.2));
const vec3 SUNSET_PURPLE = pow(vec3(0.77, 0.02, 0.3), vec3(2.2));
const vec3 SUNSET_TOP = pow(vec3(0.23, 0, 0.56), vec3(2.2));

vec3 getSkyColor(vec3 direction) {
    vec3 noonColor;
    {
        vec3 skyColor = mix(BOTTOM_COLOR, TOP_COLOR, direction.y);
        skyColor = mix(FOG_COLOR, skyColor, smoothstep(0.0, 0.1, direction.y) * 0.5 + 0.5);
        noonColor = mix(GROUND_COLOR, skyColor, smoothstep(-0.05, 0.0, direction.y));
    }
    vec3 sunsetColor;
    {
        vec3 skyColor = mix(BOTTOM_COLOR, SUNSET_TOP, direction.y);
        vec3 fogColor = mix(SUNSET_PURPLE, SUNSET_COLOR, smoothstep(0.3, 0.95, dot(sunDir, direction) * 0.5 + 0.5));
        fogColor = mix(FOG_COLOR, fogColor, smoothstep(0.2, 0.6 + direction.y, dot(sunDir, direction)) * 0.5 + 0.5);
        skyColor = mix(fogColor, skyColor, smoothstep(0.0, 0.6, direction.y) * 0.5 + 0.5);
        sunsetColor = mix(GROUND_COLOR, skyColor, smoothstep(-0.05, 0.0, direction.y));
    }
    vec3 nightColor = vec3(0, 0, 0.3);

    float time = sunDir.y / 0.9;
    vec3 color = noonColor;
    color = mix(sunsetColor, color, smoothstep(0.0, 0.1, time));
    color = mix(nightColor, color, smoothstep(-0.1, 0.0, time));

    return color;
}

const vec3 NORMAL_LIGHT_COLOR = pow(vec3(1, 0.95, 0.9), vec3(2.2));
const vec3 SUNSET_LIGHT_COLOR = pow(vec3(1, 0.51, 0.15), vec3(2.2));
const vec3 NIGHT_LIGHT_COLOR = pow(vec3(0.1, 0.1, 0.1), vec3(2.2));

vec3 getLightColor() {
    float time = sunDir.y;
    vec3 color = NORMAL_LIGHT_COLOR;
    color = mix(SUNSET_LIGHT_COLOR, color, smoothstep(0.0, 0.2, time));
    color = mix(NIGHT_LIGHT_COLOR, color, smoothstep(-0.2, 0.0, time));

    return color;
}

PathTraceOutput pathtrace(vec3 playerPos, vec3 normal) {
    vec3 rayDir = normalize(playerPos);
    vec3 blockPos = playerPos - fract(chunkOffset);

    float dist = length(playerPos);
    float offset = 0.0005 * dist * dist + 0.001;
    Ray initialRay = Ray(blockPos - rayDir * 1.0, ivec3(floor(blockPos - normal * offset)), rayDir);
    Intersection intersection = checkIntersection(initialRay);
    Material material;
    if (intersection.hit) {
        material = getMaterial(intersection);    
    } else {
        material = defaultMaterial(normal);
        material.normal = normal;
        material.albedo = pow(texture(DiffuseSampler, texCoord).rgb, vec3(GAMMA_CORRECTION));
    }
    
    vec3 radiance = vec3(0);
    vec3 throughput = vec3(1.0);
    
    vec3 origin = blockPos + normal * 0.01;
    ivec3 cell = ivec3(floor(blockPos + normal * 0.01));
    
    vec3 sunDir;
    if (checkSun(origin, cell, 55, material.normal, sunDir)) {
        vec3 brdf = brdfDirect(-rayDir, sunDir, material);
        radiance += getLightColor() * SUN_INTENSITY * brdf;
    }        

    rayDir = getNextRayDirection(material, rayDir, throughput);
    throughput = vec3(1.0);
    Ray ray = Ray(origin, cell, rayDir);

    MaterialMaskData maskData = MaterialMaskData(
        material.metal,
        material.emission,
        material.normal
    );
    
    for (int i = 0; i < 3; i++) {
        Hit hit = raytrace(ray, true, i < 2 ? 55 : 55);

        if (!hit.hit) {
            radiance += getSkyColor(ray.direction) * throughput;
            break;
        }

        // Emission
        radiance += hit.material.emission * hit.material.albedo * throughput * EMISSION_STRENGTH;

        // Sun
        vec3 sunDir;
        if (checkSun(ray.origin, ray.cell, i < 1 ? 55 : 25, hit.material.normal, sunDir)) {
            vec3 brdf = brdfDirect(-ray.direction, sunDir, hit.material);
            radiance += throughput * getLightColor() * SUN_INTENSITY * brdf;
        }
        
        // Indirect
        vec3 nextRayDir = getNextRayDirection(hit.material, ray.direction, throughput);
        
        if (!material.metal) {
            throughput *= material.ambientOcclusion;
        }

        float luma = dot(throughput, vec3(1.0)) / 3.0;
        if (randFloat() > luma) {
            break;
        }
        throughput /= luma;
        
        ray.direction = nextRayDir;
    }
    return PathTraceOutput(radiance, maskData);
}

float fresnel(float cosTheta, float F0) {
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}

void main() {
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    initRNG(uvec2(fragCoord), uvec2(InSize), frame * 1000u);
    
    float depth = texture(DepthSampler, texCoord).r;

    vec3 screenPos = vec3(texCoord, depth);
    vec3 viewPos = screenToView(screenPos, projInv);
    vec3 playerPos = viewPos * modelViewMat;
    float solidDepth;
    vec3 albedo = pow(texture(DiffuseSampler, texCoord).rgb, vec3(GAMMA_CORRECTION));

    if (depth == 1.0) {
        vec3 rayDir = normalize(playerPos);
        vec3 color = getSkyColor(rayDir);

        if (sunDir.y > -0.2) {
            color = mix(color, getLightColor() * SUN_INTENSITY, smoothstep(0.999, 0.9999, dot(sunDir, rayDir)));
        }
        color = mix(color, vec3(0.8), smoothstep(0.9995, 0.9996, dot(-sunDir, rayDir)));

        fragColor = encodeHDRColor(color);  
        return;
    }

    vec3 normal =  normalize(
        texture(NormalSampler, texCoord).rgb * 2.0 - 1.0
    );
    
    PathTraceOutput ptOutput = pathtrace(playerPos, normal);

    gl_FragDepth = storeMaterialMask(ptOutput.materialMask);
    
    fragColor = encodeHDRColor(ptOutput.radiance);
}