#version 420

#ifndef CONSTANTS_GLSL
#define CONSTANTS_GLSL

// Application specific
const ivec2 GRID_SIZE = ivec2(1024, 705);
const int AREA_SIDE_LENGTH = int(pow(float(GRID_SIZE.x * GRID_SIZE.y / 2), 1.0 / 3.0));
const int STORAGE_SIZE = 839;
const uint MAX_COUNTER = 255u;

// Universal
const float EPSILON = 0.001;
const float PI = 3.141592654;

// Temporal

const float MAX_TEMPORAL_BLENDING = 0.05;
const uint SDF_GENERATION_FRAMES = 6u;
const float MAX_SDF_DIST = 255.0;

#endif // CONSTANTS_GLSL
#ifndef RENDERSCALE_GLSL
#define RENDERSCALE_GLSL

float getRenderScale(sampler2D dataSampler) {
    return clamp(texelFetch(dataSampler, ivec2(3, 0), 0).r * 0.5 + 0.5, 0.5, 1.0);
}

vec4 scaleClipPos(vec4 clipPos, float renderScale) {
    return vec4((clipPos.xy + 1.0) * renderScale - 1.0, clipPos.zw);
}

#endif // RENDERSCALE_GLSL
#define VERTEX
const int NUMCONTROLS = 26;
const float THRESH = 0.5;
const float PROJNEAR = 0.05;
const float FPRECISION = 4000000.0;


/*
Control Map:
[0] sunDir.x
[1] sunDir.y
[2] sunDir.z
[3] arctan(ProjMat[0][0])
[4] arctan(ProjMat[1][1])
[5] ProjMat[1][0]
[6] ProjMat[0][1]
[7] ProjMat[1][2]
[8] ProjMat[1][3]
[9] ProjMat[2][0]
[10] ProjMat[2][1]
[11] ProjMat[2][2]
[12] ProjMat[2][3]
[13] ProjMat[3][0]
[14] ProjMat[3][1]
[15] ProjMat[3][2]
[16] ModelViewMat[0][0]
[17] ModelViewMat[0][1]
[18] ModelViewMat[0][2]
[19] ModelViewMat[1][0]
[20] ModelViewMat[1][1]
[21] ModelViewMat[1][2]
[22] ModelViewMat[2][0]
[23] ModelViewMat[2][1]
[24] ModelViewMat[2][2]
[25] FogColor
*/

ivec2 getScreenSize(vec2 fragCoord, vec4 glpos) {
    return ivec2(round(fragCoord * 2.0 / (glpos.xy / glpos.w + 1.0)));
}

// returns control pixel index or -1 if not control
int inControl(vec2 screenCoord, float screenWidth) {
    if (screenCoord.y < 1.0) {
        float index = floor(screenWidth / 2.0) + THRESH / 2.0;
        index = (screenCoord.x - index) / 2.0;
        if (fract(index) < THRESH && index < NUMCONTROLS && index >= 0) {
            return int(index);
        }
    }
    return -1;
}

#ifndef VERTEX

// discards the current pixel if it is control
void discardControl(vec2 screenCoord, float screenWidth) {
    if (screenCoord.y < 1.0) {
        float index = floor(screenWidth / 2.0) + THRESH / 2.0;
        index = (screenCoord.x - index) / 2.0;
        if (fract(index) < THRESH && index < NUMCONTROLS && index >= 0) {
            discard;
        }
    }
}

// discard but for when ScreenSize is not given
void discardControlGLPos(vec2 screenCoord, vec4 glpos) {
    if (screenCoord.y < 1.0) {
        float screenWidth = round(screenCoord.x * 2.0 / (glpos.x / glpos.w + 1.0));
        float index = floor(screenWidth / 2.0) + THRESH / 2.0;
        index = (screenCoord.x - index) / 2.0;
        if (fract(index) < THRESH && index < NUMCONTROLS && index >= 0) {
            discard;
        }
    }
}

#endif


// get screen coordinates of a particular control index
vec2 getControl(int index, vec2 screenSize) {
    return vec2(floor(screenSize.x / 2.0) + float(index) * 2.0 + 0.5, 0.5) / screenSize;
}

int intmod(int i, int base) {
    return i - (i / base * base);
}

vec3 encodeInt(int i) {
    int s = int(i < 0) * 128;
    i = abs(i);
    int r = intmod(i, 256);
    i = i / 256;
    int g = intmod(i, 256);
    i = i / 256;
    int b = intmod(i, 128);
    return vec3(float(r) / 255.0, float(g) / 255.0, float(b + s) / 255.0);
}

int decodeInt(vec3 ivec) {
    ivec *= 255.0;
    int s = ivec.b >= 128.0 ? -1 : 1;
    return s * (int(ivec.r) + int(ivec.g) * 256 + (int(ivec.b) - 64 + s * 64) * 256 * 256);
}

vec3 encodeFloat(float i) {
    return encodeInt(int(i * FPRECISION));
}

float decodeFloat(vec3 ivec) {
    return decodeInt(ivec) / FPRECISION;
}

uint readCounter(sampler2D counterSampler) {
    uvec4 raw = uvec4(texelFetch(counterSampler, ivec2(0), 0) * 255.0);
    return (raw.z << 16u) | (raw.y << 8u) | raw.x;
}

#undef VERTEX

#line 9 8478

in vec4 Position;

uniform vec2 InSize;
uniform sampler2D DataSampler;
uniform sampler2D CounterSampler;
uniform sampler2D SunSampler;
uniform sampler2D Atlas;

out vec2 texCoord;
out vec3 sunDir;
out mat4 projMat;
out mat3 modelViewMat;
out vec3 chunkOffset;
out vec3 rayDir;
out float near;
out float far;
out mat4 projInv;
flat out uint frame;
out float renderScale;
out vec3 steveDirection;
flat out ivec2 atlasSize;

vec4[] OFFSETS = vec4[](
    vec4(-1, -1, 0, 1),
    vec4(1, -1, 0, 1),
    vec4(1, 1, 0, 1),
    vec4(-1, 1, 0, 1)
);

void main() {
    vec4 outPos = OFFSETS[gl_VertexID];
    texCoord = outPos.xy * 0.5 + 0.5;
    
    vec2 start = getControl(0, InSize);
    vec2 inc = vec2(2.0 / InSize.x, 0.0);
    
    renderScale = getRenderScale(DataSampler);
    gl_Position = scaleClipPos(outPos, renderScale);

    projMat = mat4(
        tan(decodeFloat(texture(DataSampler, start + 3.0 * inc).xyz)), decodeFloat(texture(DataSampler, start + 6.0 * inc).xyz), 0.0, 0.0,
        decodeFloat(texture(DataSampler, start + 5.0 * inc).xyz), tan(decodeFloat(texture(DataSampler, start + 4.0 * inc).xyz)), decodeFloat(texture(DataSampler, start + 7.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 8.0 * inc).xyz),
        decodeFloat(texture(DataSampler, start + 9.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 10.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 11.0 * inc).xyz),  decodeFloat(texture(DataSampler, start + 12.0 * inc).xyz),
        decodeFloat(texture(DataSampler, start + 13.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 14.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 15.0 * inc).xyz), 0.0
    );

    modelViewMat = mat3(
        decodeFloat(texture(DataSampler, start + 16.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 17.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 18.0 * inc).xyz),
        decodeFloat(texture(DataSampler, start + 19.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 20.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 21.0 * inc).xyz),
        decodeFloat(texture(DataSampler, start + 22.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 23.0 * inc).xyz), decodeFloat(texture(DataSampler, start + 24.0 * inc).xyz)
    );
    
    vec3 forward = transpose(modelViewMat)[2];
    vec3 up = transpose(modelViewMat)[1];
    if (forward.y > 0.99) {
        steveDirection = up;
    } else {
        steveDirection = normalize(vec3(forward.x, 0.0, forward.z));
    }

    near = PROJNEAR;
    far = projMat[3][2] * near / (projMat[3][2] + 2.0 * near);

    chunkOffset = vec3(
        decodeFloat(texelFetch(DataSampler, ivec2(0, 0), 0).xyz),
        decodeFloat(texelFetch(DataSampler, ivec2(1, 0), 0).xyz),
        decodeFloat(texelFetch(DataSampler, ivec2(2, 0), 0).xyz)
    ) * 16.0;

    sunDir = vec3(
        decodeFloat(texture(DataSampler, start).xyz),
        decodeFloat(texture(DataSampler, start + inc).xyz),
        decodeFloat(texture(DataSampler, start + 2.0 * inc).xyz)
    ) * mat3(modelViewMat);
    
    float time = atan(sunDir.y, sunDir.x) / PI / 2.0 - 1.0 / 24.0;

    sunDir = normalize(sunDir);

    projInv = inverse(projMat);
    rayDir = (projInv * vec4(outPos.xy * (far - near), far + near, far - near)).xyz;
    frame = readCounter(CounterSampler);

    atlasSize = textureSize(Atlas, 0) / 2;
}