#version 420

#ifndef VOXELIZATION
#define VOXELIZATION

#ifndef CONSTANTS_GLSL
#define CONSTANTS_GLSL

// Application specific
const ivec2 GRID_SIZE = ivec2(1024, 705);
const int AREA_SIDE_LENGTH = int(pow(float(GRID_SIZE.x * GRID_SIZE.y / 2), 1.0 / 3.0));
const int STORAGE_SIZE = 839;
const uint MAX_COUNTER = 255u;

// Universal
const float EPSILON = 0.001;
const float PI = 3.141592654;

// Temporal

const float MAX_TEMPORAL_BLENDING = 0.05;
const uint SDF_GENERATION_FRAMES = 6u;
const float MAX_SDF_DIST = 255.0;

#endif // CONSTANTS_GLSL

ivec2 positionToCellScreen(vec3 position, out bool inside) {
    ivec3 sides = ivec3(AREA_SIDE_LENGTH);

    ivec3 iPosition = ivec3(floor(position));
    iPosition += sides / 2;

    inside = true;
    if (clamp(iPosition, ivec3(0), sides - 1) != iPosition) {
        inside = false;
        return ivec2(-1);
    }

    int index = (iPosition.y * sides.z + iPosition.z) * sides.x + iPosition.x;
    
    int halfWidth = GRID_SIZE.x / 2;
    ivec2 result = ivec2(
        (index % halfWidth) * 2,
        index / halfWidth + 1
    );
    result.x += result.y % 2;

    return result;
}

ivec2 cellToPixelScreen(ivec2 cell, ivec2 screenSize) {
    return ivec2(round(vec2(cell) / GRID_SIZE * screenSize));
}

ivec2 positionToPixel(vec3 position, out bool inside) {
    ivec3 sides = ivec3(AREA_SIDE_LENGTH);

    ivec3 iPosition = ivec3(floor(position));
    iPosition += sides / 2;

    inside = true;
    if (clamp(iPosition, ivec3(0), sides - 1) != iPosition) {
        inside = false;
        return ivec2(-1);
    }

    int index = (iPosition.y * sides.z + iPosition.z) * sides.x + iPosition.x;
    
    return ivec2(
        index % STORAGE_SIZE,
        index / STORAGE_SIZE
    );
}

ivec3 pixelToPosition(ivec2 pixel) {
    ivec3 sides = ivec3(AREA_SIDE_LENGTH);

    int index = pixel.x + pixel.y * STORAGE_SIZE;
    return ivec3(
        index % sides.x,
        index / sides.x / sides.z,
        (index / sides.x) % sides.z
    ) - sides / 2;
}

#endif // VOXELIZATION
#ifndef CONSTANTS_GLSL
#define CONSTANTS_GLSL

// Application specific
const ivec2 GRID_SIZE = ivec2(1024, 705);
const int AREA_SIDE_LENGTH = int(pow(float(GRID_SIZE.x * GRID_SIZE.y / 2), 1.0 / 3.0));
const int STORAGE_SIZE = 839;
const uint MAX_COUNTER = 255u;

// Universal
const float EPSILON = 0.001;
const float PI = 3.141592654;

// Temporal

const float MAX_TEMPORAL_BLENDING = 0.05;
const uint SDF_GENERATION_FRAMES = 6u;
const float MAX_SDF_DIST = 255.0;

#endif // CONSTANTS_GLSL

#line 6 98723

uniform sampler2D PrevVoxels;
uniform sampler2D PrevVoxelsDepth;
uniform sampler2D DiffuseSampler;
uniform sampler2D DepthSampler;
uniform vec2 InSize;
uniform vec2 OutSize;

in vec2 texCoord;
flat in int regenerate;
flat in ivec3 offset;

out vec4 fragColor;

const ivec3[] OFFSETS = ivec3[](
    ivec3(-1, 0, 0),
    ivec3(1, 0, 0),
    ivec3(0, -1, 0),
    ivec3(0, 1, 0),
    ivec3(0, 0, -1),
    ivec3(0, 0, 1)
);

void main() {
    ivec2 pixel = ivec2(gl_FragCoord.xy);
    ivec3 position = pixelToPosition(pixel);
    bool inside;

    if (regenerate == 1) {
        ivec2 oldPixel = cellToPixelScreen(positionToCellScreen(vec3(position), inside), ivec2(InSize));

        if (texelFetch(DepthSampler, oldPixel, 0).r != 0.0) {
            gl_FragDepth = 1.0;
            fragColor = vec4(0);
            return;
        }
        
        fragColor = texelFetch(DiffuseSampler, oldPixel, 0);
        gl_FragDepth = 0.0;
        return;
    }

    position += offset;

    ivec2 centerPixel = positionToPixel(vec3(position), inside);
    if (!inside) {
        gl_FragDepth = 1.0;
        return;
    }

    fragColor = texelFetch(PrevVoxels, centerPixel, 0);
    float dist = texelFetch(PrevVoxelsDepth, centerPixel, 0).r * MAX_SDF_DIST;
    if (dist == 0.0) {
        // Early exit when a voxel occupies the center
        gl_FragDepth = 0.0;
        return;
    }

    for (int i = 0; i < OFFSETS.length(); i++) {
        ivec2 oldPixel = positionToPixel(vec3(position + OFFSETS[i]), inside);

        if (!inside) {
            continue;
        }
        dist = min(dist, texelFetch(PrevVoxelsDepth, oldPixel, 0).r * MAX_SDF_DIST + 1.0);
    }

    gl_FragDepth = dist / MAX_SDF_DIST;
}
