#version 420

#ifndef HDR_GLSL
#define HDR_GLSL

const float MAX_BRIGHTNESS = 30.0;
const float BRIGHTNESS_CURVE = 0.1;

vec4 encodeHDRColor(vec3 color) {
    color /= MAX_BRIGHTNESS;
    color = clamp(color, 0.0, 1.0);
    color = pow(color, vec3(BRIGHTNESS_CURVE));
    color *= vec3(2047.0, 2047.0, 1023.0);
    uvec3 rawOutput = uvec3(round(color)) << uvec3(0, 11, 22);
    uint result = rawOutput.x | rawOutput.y | rawOutput.z;
    return vec4(
        result & 255u,
        (result >> 8u) & 255u,
        (result >> 16u) & 255u,
        result >> 24u
    ) / 255.0;
}

vec3 decodeHDRColor(vec4 raw) {
    uvec4 scaled = uvec4(raw * 255.0) << uvec4(0, 8, 16, 24);
    uint data = scaled.x | scaled.y | scaled.z | scaled.w;
    vec3 color = vec3(
        float(data & 2047u),
        float((data >> 11u) & 2047u),
        float(data >> 22u)
    );
    color /= vec3(2047.0, 2047.0, 1023.0);
    color = pow(color, vec3(1.0 / BRIGHTNESS_CURVE));
    return color * MAX_BRIGHTNESS;
}

float luminance(vec3 color) {
    return dot(color, vec3(0.2125, 0.7154, 0.0721));
}

#endif // HDR_GLSL
const float MAX_VARIANCE = 256.0;

float storeVariance(float value) {
    return log2(value + 1.0) / log2(MAX_VARIANCE);
}

float loadVariance(float value) {
    return exp2(value * log2(MAX_VARIANCE)) - 1.0;
}
#line 1 1010

struct MaterialMaskData {
    bool metal;
    float emission;
    vec3 normal;
};

const float MAX_VALUE = 16777215.0;

vec3 getMaskNormal(int data) {
    vec3 normal = vec3(
        (data >> 9) & 15, 
        (data >> 13) & 15, 
        0.0
    ) / 15.0 * 2.0 - 1.0;

    normal.z = sqrt(1.0 - min(1.0, dot(normal.xy, normal.xy)));
    normal.z *= float((data >> 17) & 1) * 2.0 - 1.0;    
    return normalize(normal);
}

vec3 getMaskNormal(float raw) {
    int data = int(round(raw * MAX_VALUE));
    return getMaskNormal(data);
}

MaterialMaskData getMaterialMask(float raw) {
    int data = int(round(raw * MAX_VALUE));

    return MaterialMaskData(
        (data & 1) == 1,
        ((data >> 1) & 255) / 255.0,
        getMaskNormal(data)
    );
}

float storeMaterialMask(MaterialMaskData data) {
    vec3 unitNormal = data.normal * 0.5 + 0.5;
    return (
        int(data.metal) | 
        (int(data.emission * 255.0) << 1) |
        (int(round(unitNormal.x * 15.0)) << 9) |
        (int(round(unitNormal.y * 15.0)) << 13) |
        (data.normal.z > 0.0 ? 1 : 0) << 17
    ) / MAX_VALUE;
}
#ifndef CONVERSIONS
#define CONVERSIONS

vec3 screenToView(vec3 screenPos, mat4 projInv) {
    vec4 ndc = vec4(screenPos * 2.0 - 1.0, 1.0);
    vec4 viewPos = projInv * ndc;
    return viewPos.xyz / viewPos.w;
}

vec3 viewToScreen(vec3 viewPos, mat4 projMat) {
    vec4 clipEnd = projMat * vec4(viewPos, 1);
    return clipEnd.xyz / clipEnd.w * 0.5 + 0.5;
}

float linearizeDepth(float depth, float near, float far) {
    return (near * far) / (depth * (near - far) + far);
}

#endif // CONVERSIONS
#line 7 1234

const float[] KERNEL = float[](
    3.0 / 8.0,
    1.0 / 4.0,
    1.0 / 16.0
);

const float[] GAUSS_KERNEL = float[](
    1.0 / 2.0,
    1.0 / 4.0
);

const float NORMAL_MULTIPLIER = 8.0;
const float DEPTH_MULTIPLIER = 1.0;
const float LUMA_MULTIPLIER = 80.0;

in vec2 texCoord;
in float renderScale;
in float near;
in float far;

uniform sampler2D DiffuseSampler;
uniform sampler2D DepthSampler;
uniform sampler2D NormalsSampler;
uniform sampler2D VarianceSampler;
uniform sampler2D MaterialMask;

uniform vec2 InSize;
uniform float Step;
uniform vec2 Direction;

out vec4 fragColor;

float depthToClip(float depth) {
    return (linearizeDepth(depth, near, far) - near) / (far - near) * 2.0 - 1.0;
}

void main() {
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    int stepSize = int(Step);
    
    MaterialMaskData maskData = getMaterialMask(texture(MaterialMask, texCoord * renderScale).x);
    vec3 centerNormal = texture(NormalsSampler, texCoord).rgb * 2.0 - 1.0;//maskData.normal;
    // vec3 centerNormal = maskData.normal;
    float centerDepth = texture(DepthSampler, texCoord).x;

    if (centerDepth == 1.0) {
        fragColor = texture(DiffuseSampler, texCoord * renderScale, 0);
        return;
    } 
    centerDepth = depthToClip(centerDepth);

    float centerVariance = 0.0;
    
    float gaussianTotalWeight = 0.0;
    for (int x = -1; x <= 1; x++) {
        for (int y = -1; y <= 1; y++) {
            float kernelWeight = GAUSS_KERNEL[abs(x)] * GAUSS_KERNEL[abs(y)];
            vec2 sampleCoord = vec2(gl_FragCoord.xy + vec2(x, y)) / InSize;
            if (clamp(sampleCoord, 0.0, renderScale) != sampleCoord)
                continue;
                
            centerVariance += loadVariance(texture(VarianceSampler, sampleCoord).x) * kernelWeight;
            gaussianTotalWeight += kernelWeight;
        }
    }
    float stdDeviation = sqrt(max(centerVariance / gaussianTotalWeight, 0.0));

    float centerLuma = luminance(decodeHDRColor(texture(DiffuseSampler, texCoord * renderScale)));
    float depthGradient = Direction.x > 0.5 ? dFdx(centerDepth) : dFdy(centerDepth);

    float totalWeight = 0.0;
    vec3 color = vec3(0.0);
    float newVariance = 0.0;
    int radius = maskData.metal ? (Step > 16.0 ? 0 : 1) : 2;
    for (int i = -radius; i <= radius; i++) {
        float kernelWeight = KERNEL[abs(i)];
        ivec2 off = ivec2(i * stepSize * Direction);
        vec2 sampleCoord = (vec2(fragCoord + off) + 0.5) / InSize;
        if (clamp(sampleCoord, 0.0, renderScale - 0.001) != sampleCoord)
            continue;

        float depth = texture(DepthSampler, sampleCoord / renderScale).x;
        if (depth == 1.0)
            continue;
            
        depth = depthToClip(depth); 
        vec3 normal = texture(NormalsSampler, sampleCoord / renderScale).rgb * 2.0 - 1.0;//getMaskNormal(texture(MaterialMask, sampleCoord).x);
        // vec3 normal = getMaskNormal(texture(MaterialMask, sampleCoord).x);
        float variance = loadVariance(texture(VarianceSampler, sampleCoord).x);
        vec3 sampl = decodeHDRColor(texture(DiffuseSampler, sampleCoord));
        float luma = luminance(sampl);

        float normalWeight = pow(max(0.0, dot(centerNormal, normal)), NORMAL_MULTIPLIER);
        float depthWeight = -abs(centerDepth - depth) / (DEPTH_MULTIPLIER * abs(i * stepSize * depthGradient) + 0.001);
        float lumaWeight = -abs(luma - centerLuma) / (LUMA_MULTIPLIER * stdDeviation + 0.001);

        float weight = kernelWeight * normalWeight * exp(depthWeight + lumaWeight);
        totalWeight += weight;
        color += sampl * weight;
        newVariance += variance * weight * weight;
    }
    color /= totalWeight;
    newVariance /= totalWeight * totalWeight;
    fragColor = encodeHDRColor(color); 
    gl_FragDepth = storeVariance(newVariance);
}