#version 420

#ifndef HDR_GLSL
#define HDR_GLSL

const float MAX_BRIGHTNESS = 30.0;
const float BRIGHTNESS_CURVE = 0.1;

vec4 encodeHDRColor(vec3 color) {
    color /= MAX_BRIGHTNESS;
    color = clamp(color, 0.0, 1.0);
    color = pow(color, vec3(BRIGHTNESS_CURVE));
    color *= vec3(2047.0, 2047.0, 1023.0);
    uvec3 rawOutput = uvec3(round(color)) << uvec3(0, 11, 22);
    uint result = rawOutput.x | rawOutput.y | rawOutput.z;
    return vec4(
        result & 255u,
        (result >> 8u) & 255u,
        (result >> 16u) & 255u,
        result >> 24u
    ) / 255.0;
}

vec3 decodeHDRColor(vec4 raw) {
    uvec4 scaled = uvec4(raw * 255.0) << uvec4(0, 8, 16, 24);
    uint data = scaled.x | scaled.y | scaled.z | scaled.w;
    vec3 color = vec3(
        float(data & 2047u),
        float((data >> 11u) & 2047u),
        float(data >> 22u)
    );
    color /= vec3(2047.0, 2047.0, 1023.0);
    color = pow(color, vec3(1.0 / BRIGHTNESS_CURVE));
    return color * MAX_BRIGHTNESS;
}

float luminance(vec3 color) {
    return dot(color, vec3(0.2125, 0.7154, 0.0721));
}

#endif // HDR_GLSL
const float MAX_VARIANCE = 256.0;

float storeVariance(float value) {
    return log2(value + 1.0) / log2(MAX_VARIANCE);
}

float loadVariance(float value) {
    return exp2(value * log2(MAX_VARIANCE)) - 1.0;
}
#ifndef CONSTANTS_GLSL
#define CONSTANTS_GLSL

// Application specific
const ivec2 GRID_SIZE = ivec2(1024, 705);
const int AREA_SIDE_LENGTH = int(pow(float(GRID_SIZE.x * GRID_SIZE.y / 2), 1.0 / 3.0));
const int STORAGE_SIZE = 839;
const uint MAX_COUNTER = 255u;

// Universal
const float EPSILON = 0.001;
const float PI = 3.141592654;

// Temporal

const float MAX_TEMPORAL_BLENDING = 0.05;
const uint SDF_GENERATION_FRAMES = 6u;
const float MAX_SDF_DIST = 255.0;

#endif // CONSTANTS_GLSL
#ifndef CONVERSIONS
#define CONVERSIONS

vec3 screenToView(vec3 screenPos, mat4 projInv) {
    vec4 ndc = vec4(screenPos * 2.0 - 1.0, 1.0);
    vec4 viewPos = projInv * ndc;
    return viewPos.xyz / viewPos.w;
}

vec3 viewToScreen(vec3 viewPos, mat4 projMat) {
    vec4 clipEnd = projMat * vec4(viewPos, 1);
    return clipEnd.xyz / clipEnd.w * 0.5 + 0.5;
}

float linearizeDepth(float depth, float near, float far) {
    return (near * far) / (depth * (near - far) + far);
}

#endif // CONVERSIONS
#line 7 3722

in vec2 texCoord;
in mat4 projInv;
in mat3 modelViewMat;
in vec3 cameraOffset;
in mat4 prevProjMat;
in mat3 prevModelViewMat;
in float renderScale;

uniform sampler2D DiffuseSampler;
uniform sampler2D DepthSampler;
uniform sampler2D CounterSampler;
uniform sampler2D PrevMomentsSampler;
uniform vec2 InSize;

out vec4 fragColor;

void main() {
    float counter = texture(CounterSampler, texCoord * renderScale).r * MAX_COUNTER;

    float depth = texture(DepthSampler, texCoord).r;
    if (depth == 1.0) {
        return;
    }

    float luma = luminance(decodeHDRColor(texture(DiffuseSampler, texCoord * renderScale)));
    vec2 data = vec2(luma, luma * luma);
    
    if (counter > 0.5) {
        vec3 playerPos = screenToView(vec3(texCoord, depth), projInv) * modelViewMat;
        vec3 prevPlayerPos = playerPos - cameraOffset;
        vec4 prevClipPos = prevProjMat * vec4(prevModelViewMat * prevPlayerPos, 1.0);
        vec3 prevScreenPos = prevClipPos.xyz / prevClipPos.w * 0.5 + 0.5;
        
        vec2 previousData = unpackHalf2x16(packUnorm4x8(texture(PrevMomentsSampler, prevScreenPos.xy * renderScale)));
        
        data = mix(previousData, data, max(1.0 / (counter + 1.0), MAX_TEMPORAL_BLENDING));
    }
    
    fragColor = unpackUnorm4x8(packHalf2x16(data));
}