#version 420

#ifndef HDR_GLSL
#define HDR_GLSL

const float MAX_BRIGHTNESS = 30.0;
const float BRIGHTNESS_CURVE = 0.1;

vec4 encodeHDRColor(vec3 color) {
    color /= MAX_BRIGHTNESS;
    color = clamp(color, 0.0, 1.0);
    color = pow(color, vec3(BRIGHTNESS_CURVE));
    color *= vec3(2047.0, 2047.0, 1023.0);
    uvec3 rawOutput = uvec3(round(color)) << uvec3(0, 11, 22);
    uint result = rawOutput.x | rawOutput.y | rawOutput.z;
    return vec4(
        result & 255u,
        (result >> 8u) & 255u,
        (result >> 16u) & 255u,
        result >> 24u
    ) / 255.0;
}

vec3 decodeHDRColor(vec4 raw) {
    uvec4 scaled = uvec4(raw * 255.0) << uvec4(0, 8, 16, 24);
    uint data = scaled.x | scaled.y | scaled.z | scaled.w;
    vec3 color = vec3(
        float(data & 2047u),
        float((data >> 11u) & 2047u),
        float(data >> 22u)
    );
    color /= vec3(2047.0, 2047.0, 1023.0);
    color = pow(color, vec3(1.0 / BRIGHTNESS_CURVE));
    return color * MAX_BRIGHTNESS;
}

float luminance(vec3 color) {
    return dot(color, vec3(0.2125, 0.7154, 0.0721));
}

#endif // HDR_GLSL
#ifndef CONVERSIONS
#define CONVERSIONS

vec3 screenToView(vec3 screenPos, mat4 projInv) {
    vec4 ndc = vec4(screenPos * 2.0 - 1.0, 1.0);
    vec4 viewPos = projInv * ndc;
    return viewPos.xyz / viewPos.w;
}

vec3 viewToScreen(vec3 viewPos, mat4 projMat) {
    vec4 clipEnd = projMat * vec4(viewPos, 1);
    return clipEnd.xyz / clipEnd.w * 0.5 + 0.5;
}

float linearizeDepth(float depth, float near, float far) {
    return (near * far) / (depth * (near - far) + far);
}

#endif // CONVERSIONS
#ifndef CONSTANTS_GLSL
#define CONSTANTS_GLSL

// Application specific
const ivec2 GRID_SIZE = ivec2(1024, 705);
const int AREA_SIDE_LENGTH = int(pow(float(GRID_SIZE.x * GRID_SIZE.y / 2), 1.0 / 3.0));
const int STORAGE_SIZE = 839;
const uint MAX_COUNTER = 255u;

// Universal
const float EPSILON = 0.001;
const float PI = 3.141592654;

// Temporal

const float MAX_TEMPORAL_BLENDING = 0.05;
const uint SDF_GENERATION_FRAMES = 6u;
const float MAX_SDF_DIST = 255.0;

#endif // CONSTANTS_GLSL
#line 7 7583

in vec2 texCoord;
in mat3 modelViewMat;
in mat4 prevProjMat;
in mat3 prevModelViewMat;
in mat4 projInv;
in vec3 cameraOffset;
in float near;
in float far;
in float renderScale;
in float prevRenderScale;

uniform sampler2D DiffuseSampler;
uniform sampler2D DepthSampler;

uniform sampler2D PrevDiffuseSampler;
uniform sampler2D PrevDepthSampler;

uniform sampler2D NormalsSampler;
uniform sampler2D PrevNormalsSampler;

uniform sampler2D PrevCounterSampler;

uniform vec2 InSize;

out vec4 fragColor;

vec3 bilinearSample(sampler2D sampl, vec2 fragCoord) {
    ivec2 bottomLeft = ivec2(floor(fragCoord));
    vec2 fractFragCoord = fract(fragCoord);
    return mix(
        mix(
            decodeHDRColor(texelFetch(sampl, bottomLeft + ivec2(0, 0), 0)),
            decodeHDRColor(texelFetch(sampl, bottomLeft + ivec2(1, 0), 0)),
            fractFragCoord.x
        ),
        mix(
            decodeHDRColor(texelFetch(sampl, bottomLeft + ivec2(0, 1), 0)),
            decodeHDRColor(texelFetch(sampl, bottomLeft + ivec2(1, 1), 0)),
            fractFragCoord.x
        ),
        fractFragCoord.y
    );
}

void main() {
    vec4 rawColor = texture(DiffuseSampler, texCoord * renderScale);

    fragColor = rawColor;
    gl_FragDepth = 0.0;
    
    float depth = texture(DepthSampler, texCoord).r;
    if (depth == 1.0) {
        return;
    }
    vec3 playerPos = screenToView(vec3(texCoord, depth), projInv) * modelViewMat;
    vec3 prevPlayerPos = playerPos - cameraOffset;
    vec4 prevClipPos = prevProjMat * vec4(prevModelViewMat * prevPlayerPos, 1.0);
    vec3 prevScreenPos = prevClipPos.xyz / prevClipPos.w * 0.5 + 0.5;

    if (clamp(prevScreenPos.xy, 0.0, 1.0) != prevScreenPos.xy) {
        return;
    }

    vec3 normal = texture(NormalsSampler, texCoord).rgb * 2.0 - 1.0;
    vec3 prevNormal = texture(PrevNormalsSampler, prevScreenPos.xy).rgb * 2.0 - 1.0;
    
    if (dot(normal, prevNormal) < 0.3) {
        return;
    }
    
    float prevDepth = texture(PrevDepthSampler, prevScreenPos.xy).r;
    if (abs(prevScreenPos.z - prevDepth) > 0.003) {
        return;
    }
    
    uint counter = uint(texture(PrevCounterSampler, prevScreenPos.xy * prevRenderScale).r * float(MAX_COUNTER));
    counter = min(counter + 1u, MAX_COUNTER);

    vec3 color = decodeHDRColor(rawColor);
    vec3 prevColor = bilinearSample(PrevDiffuseSampler, (prevScreenPos.xy * prevRenderScale) * InSize - 0.5);
    
    fragColor = encodeHDRColor(mix(
        prevColor,
        color,
        max(1.0 / float(counter + 1), MAX_TEMPORAL_BLENDING)
    ));
    gl_FragDepth = float(counter) / float(MAX_COUNTER);
}