#version 420

const int NUMCONTROLS = 26;
const float THRESH = 0.5;
const float PROJNEAR = 0.05;
const float FPRECISION = 4000000.0;


/*
Control Map:
[0] sunDir.x
[1] sunDir.y
[2] sunDir.z
[3] arctan(ProjMat[0][0])
[4] arctan(ProjMat[1][1])
[5] ProjMat[1][0]
[6] ProjMat[0][1]
[7] ProjMat[1][2]
[8] ProjMat[1][3]
[9] ProjMat[2][0]
[10] ProjMat[2][1]
[11] ProjMat[2][2]
[12] ProjMat[2][3]
[13] ProjMat[3][0]
[14] ProjMat[3][1]
[15] ProjMat[3][2]
[16] ModelViewMat[0][0]
[17] ModelViewMat[0][1]
[18] ModelViewMat[0][2]
[19] ModelViewMat[1][0]
[20] ModelViewMat[1][1]
[21] ModelViewMat[1][2]
[22] ModelViewMat[2][0]
[23] ModelViewMat[2][1]
[24] ModelViewMat[2][2]
[25] FogColor
*/

ivec2 getScreenSize(vec2 fragCoord, vec4 glpos) {
    return ivec2(round(fragCoord * 2.0 / (glpos.xy / glpos.w + 1.0)));
}

// returns control pixel index or -1 if not control
int inControl(vec2 screenCoord, float screenWidth) {
    if (screenCoord.y < 1.0) {
        float index = floor(screenWidth / 2.0) + THRESH / 2.0;
        index = (screenCoord.x - index) / 2.0;
        if (fract(index) < THRESH && index < NUMCONTROLS && index >= 0) {
            return int(index);
        }
    }
    return -1;
}

#ifndef VERTEX

// discards the current pixel if it is control
void discardControl(vec2 screenCoord, float screenWidth) {
    if (screenCoord.y < 1.0) {
        float index = floor(screenWidth / 2.0) + THRESH / 2.0;
        index = (screenCoord.x - index) / 2.0;
        if (fract(index) < THRESH && index < NUMCONTROLS && index >= 0) {
            discard;
        }
    }
}

// discard but for when ScreenSize is not given
void discardControlGLPos(vec2 screenCoord, vec4 glpos) {
    if (screenCoord.y < 1.0) {
        float screenWidth = round(screenCoord.x * 2.0 / (glpos.x / glpos.w + 1.0));
        float index = floor(screenWidth / 2.0) + THRESH / 2.0;
        index = (screenCoord.x - index) / 2.0;
        if (fract(index) < THRESH && index < NUMCONTROLS && index >= 0) {
            discard;
        }
    }
}

#endif


// get screen coordinates of a particular control index
vec2 getControl(int index, vec2 screenSize) {
    return vec2(floor(screenSize.x / 2.0) + float(index) * 2.0 + 0.5, 0.5) / screenSize;
}

int intmod(int i, int base) {
    return i - (i / base * base);
}

vec3 encodeInt(int i) {
    int s = int(i < 0) * 128;
    i = abs(i);
    int r = intmod(i, 256);
    i = i / 256;
    int g = intmod(i, 256);
    i = i / 256;
    int b = intmod(i, 128);
    return vec3(float(r) / 255.0, float(g) / 255.0, float(b + s) / 255.0);
}

int decodeInt(vec3 ivec) {
    ivec *= 255.0;
    int s = ivec.b >= 128.0 ? -1 : 1;
    return s * (int(ivec.r) + int(ivec.g) * 256 + (int(ivec.b) - 64 + s * 64) * 256 * 256);
}

vec3 encodeFloat(float i) {
    return encodeInt(int(i * FPRECISION));
}

float decodeFloat(vec3 ivec) {
    return decodeInt(ivec) / FPRECISION;
}

uint readCounter(sampler2D counterSampler) {
    uvec4 raw = uvec4(texelFetch(counterSampler, ivec2(0), 0) * 255.0);
    return (raw.z << 16u) | (raw.y << 8u) | raw.x;
}


in vec2 texCoord;
in mat3 modelViewMat;
in vec3 chunkOffset;
in mat4 projInv;

uniform sampler2D DiffuseSampler;
uniform sampler2D CounterSampler;
uniform sampler2D DepthSampler;
uniform sampler2D VoronoiSampler;

out vec4 fragColor;

vec3 screenToView(vec3 screenPos, mat4 projInv) {
    vec4 ndc = vec4(screenPos * 2.0 - 1.0, 1.0);
    vec4 viewPos = projInv * ndc;
    return viewPos.xyz / viewPos.w;
}

void main() {
    float time = float(readCounter(CounterSampler)) / 60.0;
    vec4 color = texture(DiffuseSampler, texCoord);
    if (distance(color.gba, vec3(1)) > 0.001) {
        fragColor = vec4(0);
        return;
    }
    
    float depth = texture(DepthSampler, texCoord).r;
    vec3 screenPos = vec3(texCoord, depth);
    vec3 view = screenToView(screenPos, projInv);
    vec3 playerPos = view * modelViewMat - chunkOffset;
    
    vec3 normal;
    int normalIndex = int(round(color.r * 255.0));
    vec3 surfaceNormal = vec3(
        (normalIndex >> 4) / 16.0 * 2.0 - 1.0,
        0.0,
        (normalIndex & 15) / 16.0 * 2.0 - 1.0
    );
    surfaceNormal.y = sqrt(1.0 - dot(surfaceNormal.xz, surfaceNormal.xz));
    
    const vec2[] DIRECTIONS = vec2[](
        normalize(vec2(1.0, 1.0)),
        normalize(vec2(-1.0, -0.3)),
        normalize(vec2(0.2, -1.0)),
        normalize(vec2(-0.3, 0.7))
    );
    
    const float[] SPEEDS = float[](
        0.04,
        0.03,
        0.02,
        0.04
    );
    float height = 0.0;
    float weight = 0.02;
    for (int i = 0; i < 4; i++) {
        height += (1.0 - texture(VoronoiSampler, playerPos.xz / 2.0 + DIRECTIONS[i] * time * SPEEDS[i])[i]) * weight;
        weight /= 2.0;
    }
    if (surfaceNormal.y > 0.9) {
        normal = normalize(cross(
            dFdx(vec3(playerPos.x, height, playerPos.z)),
            dFdy(vec3(playerPos.x, height, playerPos.z))
        ));
    } else {
        normal = surfaceNormal;
    }
    fragColor = vec4(normal * 0.5 + 0.5, 1.0);
}